from __future__ import annotations
from typing import Dict, List, Set, Tuple, Callable, Any, Optional
import logging

from weconnect.weconnect_errors import ErrorEventType
from weconnect.addressable import AddressableLeaf, AddressableObject, AddressableDict


LOG = logging.getLogger("weconnect")


class RetrievalError(Exception):
    pass


class SetterError(Exception):
    pass


class ControlError(SetterError):
    pass


class AuthentificationError(Exception):
    pass


class TemporaryAuthentificationError(AuthentificationError):
    pass


class APICompatibilityError(Exception):
    pass


class APIError(Exception):
    pass

class ErrorBus:

    def __init__(self):
        self.__errorObservers: Set[Tuple[Callable[[Optional[Any], ErrorEventType], None], ErrorEventType]] = set()

    def addErrorObserver(self, observer: Callable, errortype: ErrorEventType) -> None:
        self.__errorObservers.add((observer, errortype))
        # LOG.debug('%s: Error event observer added for type: %s', self.getGlobalAddress(), errortype)

    def removeErrorObserver(self, observer: Callable, errortype: Optional[ErrorEventType] = None) -> None:
        self.__errorObservers = filter(lambda observerEntry: observerEntry[0] == observer
                                       or (errortype is not None and observerEntry[1] == errortype), self.__errorObservers)

    def getErrorObservers(self, errortype) -> List[Any]:
        return [observerEntry[0] for observerEntry in self.getErrorObserverEntries(errortype)]

    def getErrorObserverEntries(self, errortype: ErrorEventType) -> List[Any]:
        observers: Set[Tuple[Callable, ErrorEventType]] = set()
        for observerEntry in self.__errorObservers:
            observer, observertype = observerEntry
            del observer
            if errortype & observertype:
                observers.add(observerEntry)
        return observers

    def notifyError(self, element, errortype: ErrorEventType, detail: str, message: str = None) -> None:
        observers: List[Callable] = self.getErrorObservers(errortype)
        for observer in observers:
            observer(element=element, errortype=errortype, detail=detail, message=message)
        # LOG.debug('%s: Notify called for errors with type: %s for %d observers', self.getGlobalAddress(), errortype, len(observers))



"""Unit tests for Cupra service"""

from weconnect import weconnect
from weconnect.api.cupra.elements.climatization_settings import ClimatizationSettings
from weconnect.api.cupra.elements.climatization_status import ClimatizationStatus
from weconnect.api.cupra.elements.enums import ClimatizationState
from weconnect.service import Service
from weconnect.addressable import AddressableAttribute, AddressableDict, ChangeableAttribute
from weconnect.api.cupra.elements.battery_status import BatteryStatus
from weconnect.api.cupra.elements.charging_settings import ChargingSettings
from weconnect.api.cupra.elements.charging_status import ChargingStatus
from weconnect.api.cupra.elements.vehicle import Vehicle
from weconnect.api.cupra.domain import Domain


def test_vehicles_element_minimal_construction_not_charging():
    # Given
    range_value = 280.0
    charge_level = 80.0
    mileage_km = 1451.0
    charging_remaining_time = 0
    target_charge = 100
    target_temperature_k = 295.15
    climate_remaining_time = 10

    class MockFetcher:
        base_url = 'https://example.com'
        user_id = 'USERID'

        def fetchData(self, url, force):
            # url: str = f'{self.fetcher.base_url}/v2/users/{self.fetcher.user_id}/vehicles/{self.vin.value}/mycar'
            return {
                "engines": {
                    "primary": {
                        "type": "EV",
                        "fuelType": "EV",
                        "range": {"value": range_value, "unit": "Km"},
                        "level": charge_level
                    },
                    "secondary": {
                        "type": None,
                        "fuelType": None,
                        "range": None,
                        "level": None
                    }
                },
                "measurements": {
                    "mileageKm": mileage_km
                },
                "services": {
                    "charging": {
                        "status": "NotReadyForCharging",
                        "targetPct": target_charge,
                        "chargeMode": "manual",
                        "active": False,
                        "remainingTime": charging_remaining_time,
                        "progressBarPct": 100.0
                    },
                    "climatisation": {
                        "status": "Off",
                        "targetTemperatureKelvin": target_temperature_k,
                        "active": False,
                        "remainingTime": climate_remaining_time,
                        "progressBarPct": 0.0
                    }
                }
            }
    fetcher = MockFetcher()
    vehicleDict = {
        'vin': 'VIN',
        'userRole': Vehicle.User.Role.PRIMARY_USER.value,
        'enrollmentStatus': Vehicle.User.EnrollmentStatus.COMPLETED.value,
        'userRoleStatus': Vehicle.User.UserRoleStatus.ENABLED.value,
        'vehicleNickname': 'CUPRA Born',
    }
    parent: AddressableDict[str, Vehicle] = AddressableDict(localAddress='xxx', parent=None)
    vehicle = Vehicle(
        fetcher=fetcher,
        vin='VINVINVINVINVIN',
        parent=parent,
        fromDict=vehicleDict,
        fixAPI=True,
        updateCapabilities=True,
        updatePictures=False,
        selective=None,
        enableTracker=False)
    # When
    # Note: update() is called in the constructor
    # Then
    assert vehicle.vin.value == vehicleDict['vin']
    assert vehicle.role.value.value == vehicleDict['userRole']
    assert vehicle.enrollmentStatus.value.value == vehicleDict['enrollmentStatus']
    assert vehicle.userRoleStatus.value.value == vehicleDict['userRoleStatus']
    assert vehicle.nickname.value == vehicleDict['vehicleNickname']
    # We don't have these for Cupra
    # assert vehicle.model.value == vehicleDict['model']
    # assert vehicle.devicePlatform.value == vehicleDict['devicePlatform']
    # assert vehicle.brandCode.value == vehicleDict['brandCode']
    assert vehicle.domains[Domain.MEASUREMENTS.value]['mileageKm'].odometer.value == mileage_km
    charging_status: ChargingStatus = vehicle.domains[Domain.CHARGING.value]['chargingStatus']
    assert charging_status.chargingState.value == ChargingStatus.ChargingState.NOT_READY_FOR_CHARGING
    assert charging_status.chargeMode.value == ChargingStatus.ChargeMode.MANUAL
    assert charging_status.remainingChargingTimeToComplete_min.value == charging_remaining_time
    # We don't have these for Cupra
    # charging_status.chargePower_kW
    # charging_status.chargeRate_kmph
    # charging_status.chargeType
    # charging_status.chargingSettings
    battery_status: BatteryStatus = vehicle.domains[Domain.CHARGING.value]['batteryStatus']
    assert battery_status.cruisingRangeElectric_km.value == range_value
    assert battery_status.currentSOC_pct.value == charge_level

    charging_settings: ChargingSettings = vehicle.domains[Domain.CHARGING.value]['chargingSettings']
    assert charging_settings.targetSOC_pct.value == target_charge

    climatization_settings: ClimatizationSettings = vehicle.domains[Domain.CLIMATISATION.value]['climatisationSettings']
    assert climatization_settings.targetTemperature_K.value == target_temperature_k
    assert climatization_settings.targetTemperature_C.value == target_temperature_k - 273.15
    assert climatization_settings.targetTemperature_F.value == 1.8 * (target_temperature_k - 273) + 32

    climatization_status: ClimatizationStatus = vehicle.domains[Domain.CLIMATISATION.value]['climatisationStatus']
    assert climatization_status.climatisationState.value == ClimatizationState.OFF
    # Note when fixApi == True, then this should be 0 if climatisationState == ClimatizationStatus.ClimatizationState.OFF
    assert climatization_status.remainingClimatisationTime_min.value == 0


def test_vehicles_element_minimal_construction_charging():
    # Given
    range_value = 250.0
    charge_level = 78.0
    mileage_km = 1471.0
    charging_remaining_time = 355
    target_charge = 78.0
    status_charging = "Charging"

    class MockFetcher:
        base_url = 'https://example.com'
        user_id = 'USERID'

        def fetchData(self, url, force):
            # url: str = f'{self.fetcher.base_url}/v2/users/{self.fetcher.user_id}/vehicles/{self.vin.value}/mycar'
            return {
                "engines": {
                    "primary": {
                        "type": "EV",
                        "fuelType": "EV",
                        "range": {"value": range_value, "unit": "Km"},
                        "level": charge_level
                    },
                    "secondary": {
                        "type": None,
                        "fuelType": None,
                        "range": None,
                        "level": None
                    }
                },
                "measurements": {
                    "mileageKm": mileage_km
                },
                "services": {
                    "charging": {
                        "status": status_charging,
                        "targetPct": target_charge,
                        "chargeMode": "manual",
                        "active": True,
                        "remainingTime": charging_remaining_time,
                        "progressBarPct": 100.0
                    },
                    "climatisation": {
                        "status": "Off",
                        "targetTemperatureKelvin": 295.15,
                        "active": False,
                        "remainingTime": 0,
                        "progressBarPct": 0.0
                    }
                }
            }
    fetcher = MockFetcher()
    vehicleDict = {
        'vin': 'VIN',
        'userRole': Vehicle.User.Role.PRIMARY_USER.value,
        'enrollmentStatus': Vehicle.User.EnrollmentStatus.COMPLETED.value,
        'userRoleStatus': Vehicle.User.UserRoleStatus.ENABLED.value,
        'vehicleNickname': 'CUPRA Born',
    }
    parent: AddressableDict[str, Vehicle] = AddressableDict(localAddress='xxx', parent=None)
    vehicle = Vehicle(
        fetcher=fetcher,
        vin='VINVINVINVINVIN',
        parent=parent,
        fromDict=vehicleDict,
        fixAPI=True,
        updateCapabilities=True,
        updatePictures=False,
        selective=None,
        enableTracker=False)
    # When
    # Note: update() is called in the constructor
    # Then
    assert vehicle.vin.value == vehicleDict['vin']
    assert vehicle.role.value.value == vehicleDict['userRole']
    assert vehicle.enrollmentStatus.value.value == vehicleDict['enrollmentStatus']
    assert vehicle.userRoleStatus.value.value == vehicleDict['userRoleStatus']
    assert vehicle.nickname.value == vehicleDict['vehicleNickname']
    # We don't have these for Cupra
    # assert vehicle.model.value == vehicleDict['model']
    # assert vehicle.devicePlatform.value == vehicleDict['devicePlatform']
    # assert vehicle.brandCode.value == vehicleDict['brandCode']
    assert vehicle.domains[Domain.MEASUREMENTS.value]['mileageKm'].odometer.value == mileage_km
    charging_status: ChargingStatus = vehicle.domains[Domain.CHARGING.value]['chargingStatus']
    assert charging_status.chargingState.value == ChargingStatus.ChargingState.CHARGING
    assert charging_status.chargeMode.value == ChargingStatus.ChargeMode.MANUAL
    assert charging_status.remainingChargingTimeToComplete_min.value == charging_remaining_time
    # We don't have these for Cupra
    # charging_status.chargePower_kW
    # charging_status.chargeRate_kmph
    # charging_status.chargeType
    # charging_status.chargingSettings
    battery_status: BatteryStatus = vehicle.domains[Domain.CHARGING.value]['batteryStatus']
    assert battery_status.cruisingRangeElectric_km.value == range_value
    assert battery_status.currentSOC_pct.value == charge_level

    charging_settings: ChargingSettings = vehicle.domains[Domain.CHARGING.value]['chargingSettings']
    assert charging_settings.targetSOC_pct.value == target_charge


def test_construct_weconnect_like_ha_integration():
    # When
    weconnect.WeConnect(
        username="username",
        password="password",
        service=Service('MyCupra'),
        updateAfterLogin=False,
        loginOnInit=False,
    )
    # Then just make sure we did construction without an error
    assert True

def test_climate_status():
    # Given
    class MockFetcher:
        called = False
        def post(self, url, *args):
            self.called = url
    class MockVehicle:
        fetcher = MockFetcher()
        vin: AddressableAttribute[str] = AddressableAttribute(
            localAddress='vin', parent=None, value='VINVINVIN', valueType=str)
    vehicle = MockVehicle()
    climate_dict = {
    }
    climate = ClimatizationStatus(
        vehicle=vehicle,
        parent=None,
        statusId='climatisationStatus',
        fromDict=climate_dict,
    )
    # When
    climate.climatisationState.value = ClimatizationState.COOLING
    # Then
    assert climate.climatisationState.value == ClimatizationState.COOLING
    assert vehicle.fetcher.called


def test_changeable_attribute():
    # Given
    class MockFetcher:
        called = False
        def post(self, url, *args):
            self.called = url
    class MockVehicle:
        fetcher = MockFetcher()
        vin: AddressableAttribute[str] = AddressableAttribute(
            localAddress='vin', parent=None, value='VINVINVIN', valueType=str)
    vehicle = MockVehicle()
    climate_dict = {

    }
    climate = ClimatizationSettings(
        vehicle=vehicle,
        parent=None,
        statusId='climatisationSettings',
        fromDict=climate_dict,
    )
    # When
    climate.targetTemperature_K.value = 100;
    # Then
    assert climate.targetTemperature_K.value == 100
    # There doesn't seem to be a way to set temp via the api
    assert vehicle.fetcher.called == False




from __future__ import annotations
from typing import List, Tuple, Optional
import logging
import locale

from weconnect.addressable import AddressableDict, AddressableObject
from weconnect.errors import RetrievalError
from weconnect.api.vw.domain import Domain
from weconnect.api.vw.elements.vehicle import Vehicle
from weconnect.api.vw.elements.charging_station import ChargingStation
from weconnect.fetch import Fetcher


LOG = logging.getLogger("weconnect")


class VwApi:
    def __init__(self, weconnect: AddressableObject, fetcher: Fetcher, enableTracker: bool = False, fixAPI: bool = True):
        self.base_url = 'https://mobileapi.apps.emea.vwapps.io'
        self.__vehicles: AddressableDict[str, Vehicle] = AddressableDict(localAddress='vehicles', parent=weconnect)
        self.__stations: AddressableDict[str, ChargingStation] = AddressableDict(localAddress='chargingStations', parent=weconnect)
        self.__fetcher: Fetcher = fetcher
        self.__enableTracker: bool = enableTracker
        self.fixAPI: bool = fixAPI

        # Used for charging station support
        # Public api used by weconnect-mqtt
        self.latitude: Optional[float] = None
        # Public api used by weconnect-mqtt
        self.longitude: Optional[float] = None
        # Public api used by weconnect-mqtt
        self.searchRadius: Optional[int] = None
        self.market: Optional[str] = None
        self.useLocale: Optional[str] = locale.getlocale()[0]


    @property
    def vehicles(self) -> AddressableDict[str, Vehicle]:
        return self.__vehicles

    @property
    def stations(self) -> AddressableDict[str, ChargingStation]:
        return self.__stations

    def update(self, updateCapabilities: bool = True, updatePictures: bool = True, force: bool = False,
            selective: Optional[list[Domain]] = None) -> \
                Tuple[AddressableDict[str, Vehicle], AddressableDict[str, ChargingStation]]:
        self.updateVehicles(updateCapabilities=updateCapabilities, updatePictures=updatePictures, force=force, selective=selective)
        self.updateChargingStations(force=force)
        return (self.__vehicles, self.__stations)

    def updateVehicles(self, updateCapabilities: bool = True, updatePictures: bool = True, force: bool = False,  # noqa: C901
                       selective: Optional[list[Domain]] = None) -> None:
        url = f'{self.base_url}/vehicles'
        data = self.__fetcher.fetchData(url, force)
        if data is not None:
            if 'data' in data and data['data']:
                vins: List[str] = []
                for vehicleDict in data['data']:
                    if 'vin' not in vehicleDict:
                        break
                    vin: str = vehicleDict['vin']
                    vins.append(vin)
                    try:
                        if vin not in self.__vehicles:
                            vehicle = Vehicle(fetcher=self.__fetcher, vin=vin, parent=self.__vehicles, fromDict=vehicleDict, fixAPI=self.fixAPI,
                                              updateCapabilities=updateCapabilities, updatePictures=updatePictures, selective=selective,
                                              enableTracker=self.__enableTracker)
                            self.__vehicles[vin] = vehicle
                        else:
                            self.__vehicles[vin].update(fromDict=vehicleDict, updateCapabilities=updateCapabilities, updatePictures=updatePictures,
                                                        selective=selective)
                    except RetrievalError as retrievalError:
                        LOG.error('Failed to retrieve data for VIN %s: %s', vin, retrievalError)
                # delete those vins that are not anymore available
                for vin in [vin for vin in self.__vehicles if vin not in vins]:
                    del self.__vehicles[vin]

    def updateChargingStations(self, force: bool = False) -> None:  # noqa: C901 # pylint: disable=too-many-branches
        if self.latitude is not None and self.longitude is not None:
            url: str = f'{self.base_url}/charging-stations/v2?latitude={self.latitude}&longitude={self.longitude}'
            if self.market is not None:
                url += f'&market={self.market}'
            if self.useLocale is not None:
                url += f'&locale={self.useLocale}'
            if self.searchRadius is not None:
                url += f'&searchRadius={self.searchRadius}'
            if self.__fetcher.user_id is not None:
                url += f'&userId={self.__fetcher.user_id}'
            data = self.__fetcher.fetchData(url, force)
            if data is not None:
                if 'chargingStations' in data and data['chargingStations']:
                    ids: List[str] = []
                    for stationDict in data['chargingStations']:
                        if 'id' not in stationDict:
                            break
                        stationId: str = stationDict['id']
                        ids.append(stationId)
                        if stationId not in self.__stations:
                            station: ChargingStation = ChargingStation(weConnect=self, stationId=stationId, parent=self.__stations, fromDict=stationDict,
                                                                       fixAPI=self.fixAPI)
                            self.__stations[stationId] = station
                        else:
                            self.__stations[stationId].update(fromDict=stationDict)
                    # delete those vins that are not anymore available
                    for stationId in [stationId for stationId in ids if stationId not in self.__stations]:
                        del self.__stations[stationId]

    def getChargingStations(self, latitude, longitude, searchRadius=None, market=None, useLocale=None,  # noqa: C901
                            force=False) -> AddressableDict[str, ChargingStation]:
        chargingStationMap: AddressableDict[str, ChargingStation] = AddressableDict(localAddress='', parent=None)
        url: str = f'{self.base_url}/charging-stations/v2?latitude={latitude}&longitude={longitude}'
        if market is not None:
            url += f'&market={market}'
        if useLocale is not None:
            url += f'&locale={useLocale}'
        if searchRadius is not None:
            url += f'&searchRadius={searchRadius}'
        if self.__fetcher.user_id is not None:
            url += f'&userId={self.__fetcher.user_id}'
        data = self.__fetcher.fetchData(url, force)
        if data is not None:
            if 'chargingStations' in data and data['chargingStations']:
                for stationDict in data['chargingStations']:
                    if 'id' not in stationDict:
                        break
                    stationId: str = stationDict['id']
                    station: ChargingStation = ChargingStation(weConnect=self, stationId=stationId, parent=chargingStationMap, fromDict=stationDict,
                                                               fixAPI=self.fixAPI)
                    chargingStationMap[stationId] = station
        return chargingStationMap

    def setChargingStationSearchParameters(self, latitude: float, longitude: float, searchRadius: Optional[int] = None, market: Optional[str] = None,
                                           useLocale: Optional[str] = locale.getlocale()[0]) -> None:
        self.latitude = latitude
        self.longitude = longitude
        self.searchRadius = searchRadius
        self.market = market
        self.useLocale = useLocale



from enum import Enum


class Domain(Enum):
    ALL = 'all'
    ALL_CAPABLE = 'allCapable'
    USER_CAPABILITIES = 'userCapabilities'
    MEASUREMENTS = 'measurements'
    SERVICES = 'services'
    ENGINES = 'engines'

    # Standard domains from VW api
    CHARGING = 'charging'
    CLIMATISATION = 'climatisation'

    def __str__(self):
        return self.value



from enum import Enum
import logging

from weconnect.addressable import AddressableLeaf, ChangeableAttribute, AddressableAttribute, AliasChangeableAttribute
from weconnect.api.vw.elements.generic_settings import GenericSettings
from weconnect.util import celsiusToKelvin, farenheitToKelvin

LOG = logging.getLogger("weconnect")


class ClimatizationSettings(GenericSettings):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.targetTemperature_K = ChangeableAttribute(
            localAddress='targetTemperature_K', parent=self, value=None, valueType=float)
        self.targetTemperature_C = AliasChangeableAttribute(localAddress='targetTemperature_C', parent=self, value=None,
                                                            targetAttribute=self.targetTemperature_K, conversion=celsiusToKelvin, valueType=float)
        self.targetTemperature_F = AliasChangeableAttribute(localAddress='targetTemperature_F', parent=self, value=None,
                                                            targetAttribute=self.targetTemperature_K, conversion=farenheitToKelvin, valueType=float)
        self.unitInCar = AddressableAttribute(
            localAddress='unitInCar', parent=self, value=None, valueType=ClimatizationSettings.UnitInCar)
        self.climatisationWithoutExternalPower = ChangeableAttribute(
            localAddress='climatisationWithoutExternalPower', parent=self, value=None, valueType=bool)
        self.climatizationAtUnlock = ChangeableAttribute(
            localAddress='climatizationAtUnlock', parent=self, value=None, valueType=bool)
        self.windowHeatingEnabled = ChangeableAttribute(
            localAddress='windowHeatingEnabled', parent=self, value=None, valueType=bool)
        self.zoneFrontLeftEnabled = ChangeableAttribute(
            localAddress='zoneFrontLeftEnabled', parent=self, value=None, valueType=bool)
        self.zoneFrontRightEnabled = ChangeableAttribute(
            localAddress='zoneFrontRightEnabled', parent=self, value=None, valueType=bool)
        self.zoneRearLeftEnabled = ChangeableAttribute(
            localAddress='zoneRearLeftEnabled', parent=self, value=None, valueType=bool)
        self.zoneRearRightEnabled = ChangeableAttribute(
            localAddress='zoneRearRightEnabled', parent=self, value=None, valueType=bool)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

        self.targetTemperature_K.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.climatisationWithoutExternalPower.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.climatizationAtUnlock.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.windowHeatingEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneFrontLeftEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneFrontRightEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneRearLeftEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneRearRightEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)

    def update(self, fromDict, ignoreAttributes=None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Climatization settings from dict')

        if 'value' in fromDict:
            self.targetTemperature_K.fromDict(fromDict['value'], 'targetTemperature_K')
            self.targetTemperature_C.fromDict(fromDict['value'], 'targetTemperature_C')
            self.targetTemperature_F.fromDict(fromDict['value'], 'targetTemperature_F')
            self.unitInCar.fromDict(fromDict['value'], 'unitInCar')
            self.climatisationWithoutExternalPower.fromDict(fromDict['value'], 'climatisationWithoutExternalPower')
            self.climatizationAtUnlock.fromDict(fromDict['value'], 'climatizationAtUnlock')
            self.windowHeatingEnabled.fromDict(fromDict['value'], 'windowHeatingEnabled')
            self.zoneFrontLeftEnabled.fromDict(fromDict['value'], 'zoneFrontLeftEnabled')
            self.zoneFrontRightEnabled.fromDict(fromDict['value'], 'zoneFrontRightEnabled')
            self.zoneRearLeftEnabled.fromDict(fromDict['value'], 'zoneRearLeftEnabled')
            self.zoneRearRightEnabled.fromDict(fromDict['value'], 'zoneRearRightEnabled')
        else:
            self.targetTemperature_K.enabled = False
            self.targetTemperature_C.enabled = False
            self.targetTemperature_F.enabled = False
            self.unitInCar.enabled = False
            self.climatisationWithoutExternalPower.enabled = False
            self.climatizationAtUnlock.enabled = False
            self.windowHeatingEnabled.enabled = False
            self.zoneFrontLeftEnabled.enabled = False
            self.zoneFrontRightEnabled.enabled = False
            self.zoneRearLeftEnabled.enabled = False
            self.zoneRearRightEnabled.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + [
            'targetTemperature_K',
            'targetTemperature_C',
            'targetTemperature_F',
            'unitInCar',
            'climatisationWithoutExternalPower',
            'climatizationAtUnlock',
            'windowHeatingEnabled',
            'zoneFrontLeftEnabled',
            'zoneFrontRightEnabled',
            'zoneRearLeftEnabled',
            'zoneRearRightEnabled']))

    def __str__(self):  # noqa: C901
        string = super().__str__()
        if self.targetTemperature_C.enabled:
            string += f'\n\tTarget Temperature in °C: {self.targetTemperature_C.value} °C '
        if self.targetTemperature_F.enabled:
            string += f'\n\tTarget Temperature in °F: {self.targetTemperature_F.value} °F '
        if self.targetTemperature_K.enabled:
            string += f'\n\tTarget Temperature in °K: {self.targetTemperature_K.value} °K '
        if self.unitInCar.enabled:
            string += f'\n\tTemperature unit in car: {self.unitInCar.value.value}'
        if self.climatisationWithoutExternalPower.enabled:
            string += f'\n\tClimatization without external Power: {self.climatisationWithoutExternalPower.value}'
        if self.climatizationAtUnlock.enabled:
            string += f'\n\tStart climatization after unlock: {self.climatizationAtUnlock.value}'
        if self.windowHeatingEnabled.enabled:
            string += f'\n\tWindow heating: {self.windowHeatingEnabled.value}'
        if self.zoneFrontLeftEnabled.enabled:
            string += f'\n\tHeating Front Left Seat: {self.zoneFrontLeftEnabled.value}'
        if self.zoneFrontRightEnabled.enabled:
            string += f'\n\tHeating Front Right Seat: {self.zoneFrontRightEnabled.value}'
        if self.zoneRearLeftEnabled.enabled:
            string += f'\n\tHeating Rear Left Seat: {self.zoneRearLeftEnabled.value}'
        if self.zoneRearRightEnabled.enabled:
            string += f'\n\tHeating Rear Right Seat: {self.zoneRearRightEnabled.value}'
        return string

    class UnitInCar(Enum,):
        CELSIUS = 'celsius'
        FARENHEIT = 'farenheit'
        UNKNOWN = 'unknown unit'



from __future__ import annotations
from typing import Dict, List, Any, Type, Optional
from enum import Enum
import logging

from requests import codes

from weconnect.addressable import AddressableObject, AddressableAttribute, AddressableDict, AddressableList
from weconnect.api.cupra.elements.charging_settings import ChargingSettings
from weconnect.api.cupra.elements.controls import Controls
from weconnect.api.cupra.elements.generic_status import GenericStatus
from weconnect.api.cupra.elements.generic_capability import GenericCapability
from weconnect.api.cupra.elements.charging_status import ChargingStatus
from weconnect.api.cupra.elements.parking_position import ParkingPosition
from weconnect.api.cupra.elements.odometer_measurement import OdometerMeasurement
from weconnect.api.cupra.elements.error import Error
from weconnect.api.cupra.elements.helpers.request_tracker import RequestTracker
from weconnect.api.cupra.elements.battery_status import BatteryStatus
from weconnect.errors import APICompatibilityError, APIError
from weconnect.util import kelvinToCelsius, kelvinToFarenheit, toBool
from weconnect.api.cupra.domain import Domain
from weconnect.fetch import Fetcher
from weconnect.api.cupra.elements.climatization_status import ClimatizationStatus
from weconnect.api.cupra.elements.climatization_settings import ClimatizationSettings

# Cupra
from weconnect.api.cupra.elements.engine_state import EngineState

LOG: logging.Logger = logging.getLogger("weconnect")


class DomainDict(AddressableDict):
    def __init__(self, **kwargs):
        self.error: Error = Error(localAddress='error', parent=self)
        super().__init__(**kwargs)

    def updateError(self, fromDict: Dict[str, Any]):
        if 'error' in fromDict:
            self.error.update(fromDict['error'])
        else:
            self.error.reset()

    def hasError(self) -> bool:
        return self.error.enabled


class Vehicle(AddressableObject):  # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        fetcher: Fetcher,
        vin: str,
        parent: AddressableDict[str, Vehicle],
        fromDict: Dict[str, Any],
        fixAPI: bool = True,
        updateCapabilities: bool = True,
        updatePictures: bool = True,
        selective: Optional[list[Domain]] = None,
        enableTracker: bool = False
    ) -> None:
        self.fetcher: Fetcher = fetcher
        super().__init__(localAddress=vin, parent=parent)
        
        # Public API properties
        self.vin: AddressableAttribute[str] = AddressableAttribute(
            localAddress='vin', parent=self, value=None, valueType=str)
        self.role: AddressableAttribute[Vehicle.User.Role] = AddressableAttribute(
            localAddress='role', parent=self, value=None, valueType=Vehicle.User.Role)
        self.enrollmentStatus: AddressableAttribute[Vehicle.User.EnrollmentStatus] = AddressableAttribute(
            localAddress='enrollmentStatus', parent=self, value=None, valueType=Vehicle.User.EnrollmentStatus)
        self.userRoleStatus: AddressableAttribute[Vehicle.User.UserRoleStatus] = AddressableAttribute(
            localAddress='userRoleStatus', parent=self, value=None, valueType=Vehicle.User.UserRoleStatus)
        self.model: AddressableAttribute[str] = AddressableAttribute(
            localAddress='model', parent=self, value=None, valueType=str)
        self.devicePlatform: AddressableAttribute[Vehicle.DevicePlatform] = AddressableAttribute(
            localAddress='devicePlatform', parent=self, value=None, valueType=Vehicle.DevicePlatform)
        self.nickname: AddressableAttribute[str] = AddressableAttribute(
            localAddress='nickname', parent=self, value=None, valueType=str)
        self.brandCode: AddressableAttribute[Vehicle.BrandCode] = AddressableAttribute(
            localAddress='brandCode', parent=self, value=None, valueType=Vehicle.BrandCode)
        self.capabilities: AddressableDict[str, GenericCapability] = AddressableDict(
            localAddress='capabilities', parent=self)
        self.domains: AddressableDict[str, DomainDict] = AddressableDict(
            localAddress='domains', parent=self)
        self.images: AddressableAttribute[Dict[str, str]] = AddressableAttribute(
            localAddress='images', parent=self, value=None, valueType=dict)
        self.tags: AddressableAttribute[List[str]] = AddressableAttribute(
            localAddress='tags', parent=self, value=None, valueType=list)
        self.coUsers: AddressableList[Vehicle.User] = AddressableList(
            localAddress='coUsers', parent=self)
        self.controls: Controls = Controls(
            localAddress='controls', vehicle=self, parent=self)
        
        self.fixAPI: bool = fixAPI
        self.requestTracker: Optional[RequestTracker] = None
        if enableTracker:
            self.requestTracker = RequestTracker(self)

        self.update(fromDict, updateCapabilities=updateCapabilities, updatePictures=updatePictures, selective=selective)

    def enableTracker(self) -> None:
        if self.requestTracker is None:
            self.requestTracker = RequestTracker(self)

    def disableTracker(self) -> None:
        if self.requestTracker is not None:
            self.requestTracker.clear()
        self.requestTracker = None

    def statusExists(self, domain: str, status: str) -> bool:
        if domain in self.domains and status in self.domains[domain]:
            return True
        return False

    def update(  # noqa: C901  # pylint: disable=too-many-branches
        self,
        fromDict: Dict[str, Any] = {},
        updateCapabilities: bool = True,
        updatePictures: bool = True,
        force: bool = False,
        selective: Optional[list[Domain]] = None
    ) -> None:
        if fromDict is not None:
            LOG.debug('Create /update vehicle')

            self.vin.fromDict(fromDict, 'vin')
            self.role.fromDict(fromDict, 'userRole')
            self.enrollmentStatus.fromDict(fromDict, 'enrollmentStatus')
            self.userRoleStatus.fromDict(fromDict, 'userRoleStatus')
            self.model.fromDict(fromDict, 'model')
            self.devicePlatform.fromDict(fromDict, 'devicePlatform')
            self.nickname.fromDict(fromDict, 'vehicleNickname')
            self.brandCode.fromDict(fromDict, 'brandCode')

            if updateCapabilities and 'capabilities' in fromDict and fromDict['capabilities'] is not None:
                for capDict in fromDict['capabilities']:
                    if 'id' in capDict:
                        if capDict['id'] in self.capabilities:
                            self.capabilities[capDict['id']].update(fromDict=capDict)
                        else:
                            self.capabilities[capDict['id']] = GenericCapability(
                                capabilityId=capDict['id'], parent=self.capabilities, fromDict=capDict,
                                fixAPI=self.fixAPI)
                for capabilityId in [capabilityId for capabilityId in self.capabilities.keys()
                                     if capabilityId not in [capability['id']
                                     for capability in fromDict['capabilities'] if 'id' in capability]]:
                    del self.capabilities[capabilityId]
            else:
                self.capabilities.clear()
                self.capabilities.enabled = False

            if 'images' in fromDict:
                self.images.setValueWithCarTime(fromDict['images'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.images.enabled = False

            if 'tags' in fromDict:
                self.tags.setValueWithCarTime(fromDict['tags'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.tags.enabled = False

            if 'coUsers' in fromDict and fromDict['coUsers'] is not None:
                for user in fromDict['coUsers']:
                    if 'id' in user:
                        usersWithId = [x for x in self.coUsers if x.id.value == user['id']]
                        if len(usersWithId) > 0:
                            usersWithId[0].update(fromDict=user)
                        else:
                            self.coUsers.append(Vehicle.User(localAddress=str(len(self.coUsers)), parent=self.coUsers, fromDict=user))
                    else:
                        raise APICompatibilityError('User is missing id field')
                # Remove all users that are not in list anymore
                for user in [user for user in self.coUsers if user.id.value not in [x['id'] for x in fromDict['coUsers']]]:
                    self.coUsers.remove(user)
            else:
                self.coUsers.enabled = False
                self.coUsers.clear()

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['vin',
                                              'role',
                                              'enrollmentStatus',
                                              'userRoleStatus',
                                              'model',
                                              'devicePlatform',
                                              'nickname',
                                              'brandCode',
                                              'capabilities',
                                              'images',
                                              'tags',
                                              'coUsers']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        self.updateStatus(updateCapabilities=updateCapabilities, force=force, selective=selective)

    def updateStatus(self, updateCapabilities: bool = True, force: bool = False,  # noqa: C901 # pylint: disable=too-many-branches
                selective: Optional[list[Domain]] = None):

        jobKeyClassMap: Dict[Domain, Dict[str, Type[GenericStatus]]] = {
            Domain.MEASUREMENTS: {
                'mileageKm': OdometerMeasurement,
            },

            # We will map cupra values to these standard ones
            Domain.CHARGING: {
                'batteryStatus': BatteryStatus,
                'chargingStatus': ChargingStatus,
                'chargingSettings': ChargingSettings,
            },
            Domain.CLIMATISATION: {
                'climatisationStatus': ClimatizationStatus,
                'climatisationSettings': ClimatizationSettings,
                # 'windowHeatingStatus': WindowHeatingStatus,
                # 'climatisationRequestStatus': GenericRequestStatus,
                # 'climatisationSettingsRequestStatus': GenericRequestStatus,
            },

            # Cupra only
            Domain.SERVICES: {
                'charging': ChargingStatus,     # -> Domain.CHARGING chargingStatus
                'climatisation': ClimatizationStatus
            },
            Domain.ENGINES: {
                'primary': EngineState
            }
        }

        if self.vin.value is None:
            raise APIError('VIN value is not set')
        if selective is None:
            jobs = [domain.value for domain in Domain if domain != Domain.ALL and domain != Domain.ALL_CAPABLE]
        elif Domain.ALL_CAPABLE in selective:
            if self.capabilities:
                jobs = []
                for dom in [domain for domain in Domain if domain != Domain.ALL and domain != Domain.ALL_CAPABLE]:
                    if dom.value in self.capabilities and self.capabilities[dom.value].enabled and not self.capabilities[dom.value].status.enabled:
                        jobs.append(dom.value)
                if updateCapabilities:
                    jobs.append(Domain.USER_CAPABILITIES.value)
            else:
                jobs = ['all']
        elif Domain.ALL in selective:
            jobs = ['all']
        else:
            jobs = [domain.value for domain in selective]
        
        url: str = f'{self.fetcher.base_url}/v2/users/{self.fetcher.user_id}/vehicles/{self.vin.value}/mycar'
        data: Optional[Dict[str, Any]] = self.fetcher.fetchData(url, force)

        if data is not None:
            # Iterate over top-level items in data dict
            for domain, keyClassMap in jobKeyClassMap.items():
                if not updateCapabilities and domain == Domain.USER_CAPABILITIES:
                    continue
                # if domain.value in data:
                if domain.value not in self.domains:
                    self.domains[domain.value] = DomainDict(localAddress=domain.value, parent=self.domains)
                
                for key, className in keyClassMap.items():
                    if domain.value in data:
                        if key in data[domain.value]:
                            if key in self.domains[domain.value]:
                                LOG.debug('Status %s exists, updating it', key)
                                self.domains[domain.value][key].update(fromDict=data[domain.value][key])
                            else:
                                LOG.debug('Status %s does not exist, creating it', key)
                                self.domains[domain.value][key] = className(vehicle=self, parent=self.domains[domain.value], statusId=key,
                                                                            fromDict=data[domain.value][key], fixAPI=self.fixAPI)
                    # else:
                        # Make a placeholder that we can overwrite later
                        # if key in self.domains[domain.value]:
                        # self.domains[domain.value][key] = className(vehicle=self, parent=self.domains[domain.value], statusId=key,
                        #                                                     fromDict={}, fixAPI=self.fixAPI)
                        # else:
                        #     self.domains[domain.value] = { key: None }
                if domain.value in data:
                    if 'error' in data[domain.value]:
                        self.domains[domain.value].updateError(data[domain.value])

                # check that there is no additional status than the configured ones, except for "target" that we merge into
                # the known ones
                # for key, value in {key: value for key, value in data[domain.value].items()
                #                     if key not in list(keyClassMap.keys()) and key not in ['error']}.items():
                #     LOG.warning('%s: Unknown attribute %s with value %s in domain %s', self.getGlobalAddress(), key, value, domain.value)
                # else:
                #     # Make a placeholder that we can overwrite later
                #     if domain.value not in self.domains:
                #         self.domains[domain.value] = DomainDict(localAddress=domain.value, parent=self.domains)

            # check that there is no additional domain than the configured ones
            for key, value in {key: value for key, value in data.items() if key not in list([domain.value for domain in jobKeyClassMap.keys()])}.items():
                LOG.warning('%s: Unknown domain %s with value %s', self.getGlobalAddress(), key, value)

        # Controls
        self.controls.update()

        # HACK map Cupra values back to VW values
        # We need this conditional otherwise it will fail if `data is not None`
        if Domain.SERVICES.value in self.domains:
            charging_dict = self.domains[Domain.SERVICES.value]['charging'].fromDict
            engines_dict = self.domains[Domain.ENGINES.value]['primary'].fromDict
            climatization_dict = self.domains[Domain.SERVICES.value]['climatisation'].fromDict
            # Create a charging domain object
            self.domains[Domain.CHARGING.value] = DomainDict(localAddress=Domain.CHARGING.value, parent=self.domains)

            charging_settings_dict = {
                'targetSOC_pct': charging_dict['targetPct']
            }
            self.domains[Domain.CHARGING.value]['chargingSettings'] = ChargingSettings(vehicle=self,
                parent=self.domains[Domain.CHARGING.value],
                statusId='chargingSettings',
                fixAPI=self.fixAPI,
                fromDict=charging_settings_dict)
            # We also have to call update(), not just pass fromDict to constructor
            self.domains[Domain.CHARGING.value]['chargingSettings'].update(charging_settings_dict)

            # Set charging status domain key
            charging_status_dict = {
                'remainingTime': charging_dict['remainingTime'],
                'status': charging_dict['status'],
                'chargeMode': charging_dict['chargeMode'],
            }
            self.domains[Domain.CHARGING.value]['chargingStatus'] = ChargingStatus(vehicle=self,
                parent=self.domains[Domain.CHARGING.value],
                statusId='chargingStatus',
                fixAPI=self.fixAPI,
                fromDict=charging_status_dict,
            )
            # # We also have to call update(), not just pass fromDict to constructor
            self.domains[Domain.CHARGING.value]['chargingStatus'].update(charging_status_dict)

            # Set battery status domain key
            battery_status_dict = {
                'currentSOC_pct': engines_dict['level'],
                'cruisingRangeElectric_km': engines_dict['range']['value']
            }
            self.domains[Domain.CHARGING.value]['batteryStatus'] = BatteryStatus(vehicle=self,
                parent=self.domains[Domain.CHARGING.value],
                statusId='batteryStatus',
                fixAPI=self.fixAPI,
                fromDict=battery_status_dict)
            # We also have to call update(), not just pass fromDict to constructor
            self.domains[Domain.CHARGING.value]['batteryStatus'].update(battery_status_dict)

            # Climate status domain key
            climate_status_dict = {
                'climatisationState': climatization_dict['status'],
                'remainingClimatisationTime_min': climatization_dict['remainingTime']
            }
            self.domains[Domain.CLIMATISATION.value]['climatisationStatus'] = ClimatizationStatus(vehicle=self,
                parent=self.domains[Domain.CLIMATISATION.value],
                statusId='climatisationStatus',
                fixAPI=self.fixAPI,
                fromDict=climate_status_dict)
            # We also have to call update(), not just pass fromDict to constructor
            self.domains[Domain.CLIMATISATION.value]['climatisationStatus'].update(climate_status_dict)

            # Climate settings domain key
            climate_settings_dict = {
                'targetTemperature_K': climatization_dict['targetTemperatureKelvin'],
                'targetTemperature_C': kelvinToCelsius(float(climatization_dict['targetTemperatureKelvin'])),
                'targetTemperature_F': kelvinToFarenheit(float(climatization_dict['targetTemperatureKelvin']))
            }
            self.domains[Domain.CLIMATISATION.value]['climatisationSettings'] = ClimatizationSettings(vehicle=self,
                parent=self.domains[Domain.CLIMATISATION.value],
                statusId='climatisationSettings',
                fixAPI=self.fixAPI,
                fromDict=climate_settings_dict)
            # We also have to call update(), not just pass fromDict to constructor
            self.domains[Domain.CLIMATISATION.value]['climatisationSettings'].update(climate_settings_dict)

    def __str__(self) -> str:  # noqa: C901
        returnString: str = ''
        if self.vin.enabled and self.vin.value is not None:
            returnString += f'VIN:               {self.vin.value}\n'
        if self.model.enabled and self.model.value is not None:
            returnString += f'Model:             {self.model.value}\n'
        if self.devicePlatform.enabled and self.devicePlatform.value is not None:
            returnString += f'Device Platform:   {self.devicePlatform.value.value}\n'
        if self.nickname.enabled and self.nickname.value is not None:
            returnString += f'Nickname:          {self.nickname.value}\n'
        if self.brandCode.enabled and self.brandCode.value is not None:
            returnString += f'Brand Code:        {self.brandCode.value.value}\n'
        if self.role.enabled and self.role.value is not None:
            returnString += f'Role:              {self.role.value.value}\n'  # pylint: disable=no-member
        if self.enrollmentStatus.enabled and self.enrollmentStatus.value is not None:
            returnString += f'Enrollment Status: {self.enrollmentStatus.value.value}\n'  # pylint: disable=no-member
        if self.userRoleStatus.enabled and self.userRoleStatus.value is not None:
            returnString += f'User Role Status:  {self.userRoleStatus.value.value}\n'  # pylint: disable=no-member
        if self.coUsers.enabled:
            returnString += f'Co-Users: {len(self.coUsers)} items\n'
            for coUser in self.coUsers:
                if coUser.enabled:
                    returnString += ''.join(['\t' + line for line in str(coUser).splitlines(True)]) + '\n'
        if self.tags.enabled and self.tags.value:
            returnString += 'Tags:               ' + ', '.join(self.tags.value) + '\n'
        if self.capabilities.enabled:
            returnString += f'Capabilities: {len(self.capabilities)} items\n'
            for capability in self.capabilities.values():
                if capability.enabled:
                    returnString += ''.join(['\t' + line for line in str(capability).splitlines(True)]) + '\n'
        if self.domains.enabled:
            returnString += f'Domains: {len(self.domains)} items\n'
            for domain in self.domains:
                returnString += f'[{domain}] Elements: {len(self.domains[domain])} items\n'
                for status in self.domains[domain].values():
                    if status.enabled:
                        returnString += ''.join(['\t' + line for line in str(status).splitlines(True)]) + '\n'
                if self.domains[domain].hasError():
                    returnString += ''.join(['\t' + line for line in f'Error: {self.domains[domain].error}'.splitlines(True)]) + '\n'
        return returnString

    class Badge(Enum):
        CHARGING = 'charging'
        CONNECTED = 'connected'
        COOLING = 'cooling'
        HEATING = 'heating'
        LOCKED = 'locked'
        PARKING = 'parking'
        UNLOCKED = 'unlocked'
        VENTILATING = 'ventilating'
        WARNING = 'warning'

    class DevicePlatform(Enum,):
        MBB = 'MBB'
        MBB_ODP = 'MBB_ODP'
        MBB_OFFLINE = 'MBB_OFFLINE'
        WCAR = 'WCAR'
        UNKNOWN = 'UNKNOWN'

    class BrandCode(Enum,):
        V = 'V'
        UNKNOWN = 'unknown brand code'

    class User(AddressableObject):
        def __init__(
            self,
            localAddress: str,
            parent: AddressableObject,
            fromDict: Dict[str, str] = {},
        ) -> None:
            super().__init__(localAddress=localAddress, parent=parent)
            self.id: AddressableAttribute[str] = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=str)
            self.role: AddressableAttribute[Vehicle.User.Role] = AddressableAttribute(localAddress='role', parent=self, value=None, valueType=Vehicle.User.Role)
            self.roleReseted: AddressableAttribute[bool] = AddressableAttribute(localAddress='roleReseted', parent=self, value=None, valueType=bool)
            self.enrollmentStatus: AddressableAttribute[Vehicle.User.EnrollmentStatus] = AddressableAttribute(
                localAddress='enrollmentStatus', parent=self,
                value=None,
                valueType=Vehicle.User.EnrollmentStatus)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict) -> None:
            LOG.debug('Update User from dict')

            if 'id' in fromDict:
                self.id.setValueWithCarTime(fromDict['id'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.id.enabled = False

            if 'role' in fromDict and fromDict['role']:
                try:
                    self.role.setValueWithCarTime(Vehicle.User.Role(fromDict['role']), lastUpdateFromCar=None, fromServer=True)
                except ValueError:
                    self.role.setValueWithCarTime(Vehicle.User.Role.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
                    LOG.warning('An unsupported role: %s was provided, please report this as a bug', fromDict['role'])
            else:
                self.role.enabled = False

            if 'roleReseted' in fromDict:
                self.roleReseted.setValueWithCarTime(toBool(fromDict['roleReseted']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.roleReseted.enabled = False

            if 'enrollmentStatus' in fromDict and fromDict['enrollmentStatus']:
                try:
                    self.enrollmentStatus.setValueWithCarTime(Vehicle.User.EnrollmentStatus(fromDict['enrollmentStatus']), lastUpdateFromCar=None,
                                                              fromServer=True)
                except ValueError:
                    self.enrollmentStatus.setValueWithCarTime(Vehicle.User.EnrollmentStatus.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
                    LOG.warning('An unsupported target operation: %s was provided, please report this as a bug', fromDict['enrollmentStatus'])
            else:
                self.enrollmentStatus.enabled = False

        def __str__(self) -> str:
            returnValue: str = ''
            if self.id.enabled and self.id.value is not None:
                returnValue += f'Id: {self.id.value}, '
            if self.role.enabled and self.role.value is not None:
                returnValue += f' Role: {self.role.value.value}, '  # pylint: disable=no-member
            if self.roleReseted.enabled and self.roleReseted.value is not None:
                returnValue += f' Reseted: {self.roleReseted.value}, '
            if self.enrollmentStatus.enabled and self.enrollmentStatus.value is not None:
                returnValue += f' Enrollment Status: {self.enrollmentStatus.value.value}'  # pylint: disable=no-member
            return returnValue

        class Role(Enum,):
            PRIMARY_USER = 'PRIMARY_USER'
            SECONDARY_USER = 'SECONDARY_USER'
            GUEST_USER = 'GUEST_USER'
            CDIS_UNKNOWN_USER = 'CDIS_UNKNOWN_USER'
            UNKNOWN = 'UNKNOWN'

            def __str__(self) -> str:
                return self.value

        class EnrollmentStatus(Enum,):
            STARTED = 'started'
            NOT_STARTED = 'not_started'
            COMPLETED = 'completed'
            GDC_MISSING = 'gdc_missing'
            INACTIVE = 'inactive'
            UNKNOWN = 'unknown'


        class UserRoleStatus(Enum,):
            ENABLED = 'ENABLED'
            DISABLED_HMI = 'DISABLED_HMI'
            DISABLED_SPIN = 'DISABLED_SPIN'
            DISABLED_PU_SPIN_RESET = 'DISABLED_PU_SPIN_RESET'
            CDIS_UNKNOWN_USER = 'CDIS_UNKNOWN_USER'
            UNKNOWN = 'UNKNOWN'



from enum import Enum
import logging

import json
import requests

from weconnect.addressable import AddressableLeaf, ChangeableAttribute, AliasChangeableAttribute
from weconnect.api.cupra.elements.enums import ClimatizationState
from weconnect.api.cupra.elements.generic_status import GenericStatus
from weconnect.api.cupra.elements.error import Error
from weconnect.errors import SetterError
from weconnect.api.cupra.domain import Domain

LOG = logging.getLogger("weconnect")


class GenericSettings(GenericStatus):
    def valueChanged(self, element, flags):  # noqa: C901
        del element
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED \
                and not flags & AddressableLeaf.ObserverEvent.UPDATED_FROM_SERVER:
            action = self.id.partition('Settings')[0]
            if action == 'climatisationStatus':
                action = 'climatisation'
            # print(action)

            # Figure out state
            settingsDict = dict()
            for child in self.getLeafChildren():
                if isinstance(child, ChangeableAttribute) and not isinstance(child, AliasChangeableAttribute):
                    if isinstance(child.value, Enum):  # pylint: disable=no-member # this is a fales positive
                        settingsDict[child.getLocalAddress()] = child.value.value  # pylint: disable=no-member # this is a fales positive
                    else:
                        settingsDict[child.getLocalAddress()] = child.value  # pylint: disable=no-member # this is a fales positive
            # print(settingsDict)

            # if 'targetTemperature_K' in settingsDict:
            #     state = ''
            if 'climatisationState' in settingsDict:
                if settingsDict['climatisationState'] in [
                        ClimatizationState.COOLING.value,
                        ClimatizationState.HEATING.value,
                        ClimatizationState.VENTILATION.value,
                        ClimatizationState.ON.value
                    ]:
                    state = 'start'
                else:
                    state = 'stop'
            else:
                return


            url = f'https://ola.prod.code.seat.cloud.vwgroup.com/vehicles/{self.vehicle.vin.value}/{action}/requests/{state}'
            # print(url)

            self.vehicle.fetcher.post(url)
            # putResponse = self.vehicle.weConnect.session.put(url, data=data, allow_redirects=True)

            # TODO handle response
            # if putResponse.status_code != requests.codes['ok']:
            #     errorDict = putResponse.json()
            #     if errorDict is not None and 'error' in errorDict:
            #         error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
            #         if error is not None:
            #             message = ''
            #             if error.message.enabled and error.message.value is not None:
            #                 message += error.message.value
            #             if error.info.enabled and error.info.value is not None:
            #                 message += ' - ' + error.info.value
            #             if error.retry.enabled and error.retry.value is not None:
            #                 if error.retry.value:
            #                     message += ' - Please retry in a moment'
            #                 else:
            #                     message += ' - No retry possible'
            #             raise SetterError(f'Could not set value ({message})')
            #         else:
            #             raise SetterError(f'Could not set value ({putResponse.status_code})')
            #     raise SetterError(f'Could not not set value ({putResponse.status_code})')
            
            # TODO track stuff?
            # responseDict = putResponse.json()
            # if 'data' in responseDict and 'requestID' in responseDict['data']:
            #     if self.vehicle.requestTracker is not None:
            #         self.vehicle.requestTracker.trackRequest(responseDict['data']['requestID'], Domain.ALL, 20, 120)



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute
from weconnect.api.cupra.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class EngineState(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.type = AddressableAttribute(
            localAddress='type', 
            value=None, 
            parent=self, 
            valueType=EngineState.EngineType
        )
        self.fuelType = AddressableAttribute(
            localAddress='fuelType', 
            value=None, 
            parent=self, 
            valueType=EngineState.FuelType
        )
        self.level = AddressableAttribute(
            localAddress='level', 
            value=None, 
            parent=self, 
            valueType=float)
        self.range = AddressableAttribute(
            localAddress='range', 
            value=None, 
            parent=self, 
            valueType=float)
        self.rangeUnit = AddressableAttribute(
            localAddress='range', 
            value=None, 
            parent=self, 
            valueType=str)

        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []

        self.type.fromDict(fromDict, self.type.localAddress)
        self.fuelType.fromDict(fromDict, self.fuelType.localAddress)
        self.level.fromDict(fromDict, self.level.localAddress)
        self.range.setValueWithCarTime(fromDict[self.range.localAddress].get('value'))
        self.rangeUnit.setValueWithCarTime(fromDict[self.range.localAddress].get('unit'))

    def __str__(self):
        string = super().__str__()
        string += f'\n\tEngine Type: {self.type.value}'
        string += f'\n\tFuel Type: {self.fuelType.value}'
        string += f'\n\tFuel Level: {self.level.value}'
        string += f'\n\tRange: {self.range.value} {self.rangeUnit.value}'
        return string

    class EngineType(Enum,):
        EV = 'EV'

    class FuelType(Enum,):
        EV = 'EV'
